@@ -842,7 +842,8 @@ callcompfunc(char *s, char *fn)
 	}
 	startauto = ((compinsert &&
 		      !strcmp(compinsert, "automenu-unambiguous")) ||
-		     (bashlistfirst && (!compinsert || !*compinsert)));
+		     (bashlistfirst && isset(AUTOMENU) &&
+                      (!compinsert || !*compinsert)));
 	useexact = (compexact && !strcmp(compexact, "accept"));
 
 	if (!comptoend || !*comptoend)
